import sys
from typing import Optional
import secrets
import os
import random
from cryptography.fernet import Fernet
from base64 import urlsafe_b64encode as b64e, urlsafe_b64decode as b64d

from cryptography.hazmat.backends  import default_backend as backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

DEFAULT_DATABASE_PATH = "db.txt"
USAGES = {
        "new": "new <length> [ <label> <character set> ] ( 'length' = num characters in password, 'label' = optional name for the password 'charset' = characters that can be in the password (use '-<characters>' to exclude a list of characters))",
        "echo": "echo <string content> ( 'string_content' = any text to be echoed )",
        "dbg": "dbg ( Prints a debug output of the manager's state )",
        "list": "list [ <label> ] (Prints the passwords currently stored 'label' = the password you want to decrypt)",
        "save": "save - Saves the state to the file originally inputted.", # TODO: make it possible to save to a new file
        "drop": "drop <label> - deletes the selected label from the list of passwords. This can be undone with the 'undo' command, however if 'drop' is used on an already deleted password, the password is removed forever from the database.",
        "undo": "undo - undoes the latest deletion/modification",
        "history": "history - prints all the deleted passwords",
        "cfg": "config [ <param=newparam> ] - configures the current program state or lists it when run without arguments. Example: 'config hints=False'",
        "rename": "rename <label> <newlabel> - renames a password",
}
DEFAULT_KEYGEN_ITERATIONS = 100_000
LOWERCASE_ALPHABET = set([chr(e) for e in range(ord('a'), ord('a') + 25)])
UPPERCASE_ALPHABET = set([chr(e) for e in range(ord('A'), ord('A') + 25)])
NUMBERS = set([chr(e) for e in range(ord('0'), ord('0') + 10)])
SPECIAL = set([chr(e) for e in range(ord('!'), ord('!') + 14)]) ^ set(',') # no commas in our passwords, sorry
CHARSET_MAP = {
        "l": LOWERCASE_ALPHABET,
        "u": UPPERCASE_ALPHABET,
        "n": NUMBERS,
        "s": SPECIAL,
}

def confirm(msg, yes, no):
    i = input(msg + f" ({yes}/{no})")
    i.strip()
    s = i.lower()
    if s == yes:
        return True
    elif s == no:
        return False
    else:
        return False

def derive_key(password: bytes, salt: bytes, iterations: int = DEFAULT_KEYGEN_ITERATIONS) -> bytes:
    kdf = PBKDF2HMAC(
            algorithm = hashes.SHA256(), length=32, salt=salt,
            iterations=iterations, backend=backend)
    return b64e(kdf.derive(password))

def password_encrypt(message: bytes, password: str, iterations: int = DEFAULT_KEYGEN_ITERATIONS) -> bytes:
    salt = secrets.token_bytes(16)
    key = derive_key(password.encode(), salt, iterations)
    return b64e(
            b'%b%b%b' % (
                salt,
                iterations.to_bytes(4, 'big'),
                b64d(Fernet(key).encrypt(message)),
            )
    )

def password_decrypt(token: bytes, password: str) -> (bytes, bytes):
    decoded = b64d(token)
    salt, iters, token = decoded[:16], decoded[16:20], b64e(decoded[20:])
    iterations = int.from_bytes(iters, 'big')
    key = derive_key(password.encode(), salt, iterations)
    return (key, Fernet(key).decrypt(token))

def encrypt(message: bytes, key: bytes) -> bytes:
    return Fernet(key).encrypt(message)

def decrypt(token: bytes, key: bytes) -> bytes:
    return Fernet(key).decrypt(token)

class State:
    def __init__(self, filepath, password):
        self.filepath = filepath
        try:
            file = open(filepath, 'r')
        except:
            raise FileNotFoundError

        content = file.read()
        lines = content.split('\n')
        
        head = lines[0]
        try:
            key, decrypt = password_decrypt(head, password)
        except:
            print("Password does not match for this database, and as such it cannot be decrypted.")
            raise Exception

        self.passwords = {}
        self.to_remove = []
        self.overwritten = []
        self.key = key
        self.head = head

        if content == "":
            file.close()
            return

        for i, line in enumerate(lines):
            if i == 0:
                continue # lmao so hacky! whaddeva!!
            if line == "":
                continue
            parts = line.split(",")
            if len(parts) != 2:
                print(f"Error parsing password on line {i}: {line}\n ^^^^ -> Expected comma.")
                continue
            l = parts[0].split("(")
            if len(l) != 1 and l[0] == "rm":
                part = parts[0].split("(")[1][:-1]
                self.passwords[part] = parts[1].encode()
                self.to_remove.append(part)
            elif len(l) != 1 and l[0] == "old":
                self.overwritten.append(l[1][:-1])
                self.passwords[parts[0]] = parts[1].encode()
            else:
                self.passwords[parts[0]] = parts[1].encode()

        file.close()

    def add_password(self, label, password):
        if label == "":
            label = str(len(self.passwords) + 1)
        pourmoi = self.key
        self.passwords[label] = Fernet(self.key).encrypt(password.encode())
        #self.passwords[label] = encrypt(password.encode, self.key)

    def get_password(self, label):
        if label in self.passwords:
            return Fernet(self.key).decrypt(self.passwords[label])
        else:
            return b"[ERROR: no such label found]"

    def serialize_password(self, key, value):
        if key in self.to_remove:
            return f"rm({key}),{value.decode('utf-8')}"
        # elif key in self.overwritten:
        #     del self.overwritten[self.overwritten.index(key)]
        #     return f"old({key}),{value.decode('utf-8')}"
        else:
            return f"{key},{value.decode('utf-8')}"


    def save(self):
        content = self.head + "\n" + "\n".join(map(lambda x: self.serialize_password(x[0], x[1]), self.passwords.items()))
        
        file = open(self.filepath, 'w')
        file.write(content)
        file.close()


def generate_password(length, charset):
    s = []
    for i in range(0, length):
        s.append(random.choice(list(charset)))
    return ''.join(s)

def handle_args(db: str, password: str):
    state = State(db, password)
    return state

def handle_command(state: State, args: [str], cfg) -> Optional[bool]:
    command = args[0]
    if command == "q" or command == "quit" or command == "exit":
        if len(args) == 2:
            if args[1] == "f" or args[1] == "force" or args[1] == "nosave":
                if len(args) > 2 and args[2] == "y":
                    return True
                elif confirm("Are you sure you want to exit without saving?", "y", "n"):
                    return True
        else:
            return False
    if command == "cfg" or command == "config" or command == "settings":
        if len(args) < 2:
            print(f"Settings: hints: {cfg['hints']}, ...")
            return None

        s = args[1].split("=")
        if len(s) < 2:
            print(f"Incorrect format for modifying program parameters. Usage: {USAGES['cfg']}")
            return None

        if s[0] not in cfg:
            print(f"Unknown cfg parameter '{s[0]}'. Type 'cfg' for a list of config parameters.")
            return None
        try:
            cfg[s[0]] = s[1]
        except:
            print(f"Incorrect format for cfg parameter {s[0]}, expected type {type(cfg[s[0]])}")
    elif command == "help" or command == "h" or command == "?":
        if len(args) == 1:
            print("Available commands: ")
            print("\t" + ", ".join(USAGES.keys()))
            if cfg['hints'] == True: 
                print("(Hint: type 'help <command name>' for more detailed info about that command)")
        else:
            cmd = args[1]
            if not cmd in USAGES:
                print("Unknown command, you may have mistyped it.")
                return None
            print(USAGES[cmd])
    elif command == "echo":
        if len(args) < 2:
            print(f"Too few arguments for {command}, usage: {USAGES['echo']}") 
            return None
        print(" ".join(args[1:]))
    elif command == "save" or command == "s":
        try:
            state.save()
            print(f"State saved to {filepath}!")
        except:
            print("There was an error saving, maybe back up your stuff now. Sorry")
    elif command == "drop" or command == "delete" or command == "del" or command == "rm":
        if len(args) < 2:
            print(f"Too few arguments for command {command}, usage: {USAGES['drop']}")
            return None
        
        label = args[1].strip().lower()
        if label in state.to_remove:
            if confirm(f"Are you sure you want to PERMANANTLY DELETE {label}? FOREVER????? (Note: this CANNOT be undone)", "y", "n"):
                del state.passwords[label]
                del state.to_remove[state.to_remove.index(label)]
                print(f"Password '{label}' was permanently removed from database. This cannot be undone.")
        elif label in state.passwords:
            if confirm(f"Are you sure you want to delete {label}?", "y", "n"):
                state.to_remove.append(label)
                print(f"Password '{label}' removed from database.")
                if cfg['hints'] == True:
                    print("(Hint: use 'undo' to undo this operation, and 'history' to see previous deleted passwords and restore them)")
        else:
            print(f"Password '{label}' was not found, have you spelt it correctly?")
    elif command == "undo" or command == "un":
        if len(state.to_remove) == 0:
            print("Already on latest change")
            return None
        ltst = state.to_remove.pop()
        print(f"Undone latest deletion: {ltst}")
    elif command == "restore" or command == "res" or command == "r":
        if len(args) < 2:
            print(f"Too few arguments for command {command}. Usage: {USAGES['restore']}")
            return None

        args[1].strip()
        og_label = args[1].lower()
        label = og_label
        if og_label in state.passwords and (og_label in state.overwritten or og_label in state.to_remove):
            print("You will have to rename this password to restore it.")
            label_in = input("New label (If left empty, a number will be used): ")
            if label_in == "":
                label = str(len(state.passwords) + 1) # oh so hacky, i hate python and i want my damn state to be operated on by its own damn functions. I hate this.
            else:
                label_in.strip()
                label = label_in.lower()

        if og_label in state.to_remove:
            del state.to_remove[state.to_remove.index(og_label)]
            print(f"Restored {og_label}")
        elif og_label in state.overwritten:
            del state.overwritten[state.overwritten.index(og_label)]
            state.passwords[label] = state.passwords[f"old({og_label})"]
            del state.passwords[f"old({og_label})"]
            print(f"Restored {og_label}")
        else:
            print(f"Label '{og_label}' could not be found in deleted or overwritten passwords")
    elif command == "rename" or command == "rn":
        if len(args) < 3:
            print(f"Too few args for {command}. Usage: {USAGES['rename']}.")
            return None

        if args[1] in state.passwords:
            state.passwords[args[2]] = state.passwords[args[1]]
            del state.passwords[args[1]]
            print(f"{args[1]} -> {args[2]}")
        else:
            print(f"{args[1]} is not a known password label")
    elif command == "history" or command == "his" or command == "undolist":
        print("Deleted:")
        for remd in state.to_remove:
            print(f"{remd}: [...]", end=' ')
        print("\nOverwritten:")
        for over in state.overwritten:
            print(f"{over}: [...]", end=' ')
        print("")
    elif command == "new": # Form: new <length> [ <label> ]
        if len(args) < 2:
            print(f"Too few arguments for {command}, usage: {USAGES['new']}") 
            return None
        try:
            length = int(args[1])
        except:
            print(f"Error parsing length as number, usage: {USAGES['new']}")
            return None

        if len(args) >= 3:
            label = args[2]
        else:
            label = ""

        if label in state.passwords:
            print(f"Error generating new password, label '{label}' is already in use.")
            if label in state.overwritten:
                if confim("Do you want to PERMANENTLY OVERWRITE {label}? (THIS CANNOT BE UNDONE)", "y", "n"):
                    todo
            if confirm("Would you like to replace it?", "y", "n"):
                state.overwritten.append(label)
                state.passwords[f"old({label})"] = state.passwords[label]
                print(f"Overwriting '{label}' with new password...")
                if cfg['hints'] == True:
                    print("(Hint: use the 'restore' command to undo this action)")
            else:
                return None

        charset = "all"

        chars = set()

        if len(args) == 4:
            charset = args[3]

        if charset == "all":
            chars = LOWERCASE_ALPHABET | UPPERCASE_ALPHABET | NUMBERS | SPECIAL
        else:
            s = charset.split("-")
            for c in list(s[0]):
                chars |= CHARSET_MAP[c]
            if len(s) > 1:
                for c in s[1]: # list of chars to exclude
                    chars ^= set(c)

        password = generate_password(length, chars)
        print(str(password))
        state.add_password(label, password)
    elif command == "list" or command == "ls":
        if len(args) == 1:
            # if state.passwords.len() >= 10:
                # TODO: make it possible to scroll thru a list of passwords
            for l in state.passwords:
                if l not in state.to_remove and "old" not in l:
                    print(f"{l}: [...]")
            if cfg['hints'] == True:
                print("(Hint: type 'list <key>' to decrypt your chosen password")
        elif len(args) == 2:
            print(state.get_password(args[1]).decode('utf-8'))
    return None

def main():
    filepath = ""
    if len(sys.argv) < 2:
        print(f"No database path supplied. Usage: {sys.argv[0]} <database path> <password>")
        return
    else:
        filepath = sys.argv[1]
    
    if len(sys.argv) < 3:
        print(f"No password supplied. Usage: {sys.argv[0]} <database path> <password>");
        return

    password = sys.argv[2]

    try:
        state = handle_args(filepath, password)
    except FileNotFoundError:
        print(f"Error while opening database file, {filepath} does not exist or cannot be opened.")
        ask = input(f"Create {filepath}? (Y/N)")
        ask = ask.strip()
        ask = ask.lower()
        if ask == "y":
            file = open(filepath, "xb")
            salt = secrets.token_bytes(16)
            key = password_encrypt("super secret hihi >_>".encode(), password)
            file.write(key)
            file.write('\n'.encode())
            file.close()
            return main()
        else:
            return
    except Exception as e:
        print(f"Error parsing database {filepath}: {e}")
        return

    # TODO: make this serialized (somehow... :[)
    cfg = {
        "hints": True
    }
    print("Welcome to SmashPass. Type 'h' or 'help' for help")

    while True:
        command_input = input(">")
        command_input = command_input.strip()

        args = command_input.split(" ")
        if len(args) == 0:
            continue

        should_immediately_exit = None
        try:
            should_immediately_exit = handle_command(state, args, cfg)
        except e:
            print(f"ERROR: handling the inputted command: {e}")
        if should_immediately_exit is not None:
            if should_immediately_exit:
                return
            else:
                break

    state.save()
    print("Thanks for using!")


if __name__ == "__main__":
    main()
